#!/usr/bin/env dart

import 'dart:async';
import 'dart:io';
import 'dart:isolate';

import 'package:watcher/watcher.dart';

main(List<String> args) async {
  final arguments = args.toList();
  if (arguments.length < 1) return print('Usage: embla start');
  final command = arguments.removeAt(0);
  if (command != 'start') return print('''Usage: embla start <options>

Options:
  [--isolates | -i] <int>    The number of instances to run (defaults to 1)
  [--production]             Run without file watcher
''');
  final productionMode = () {
    if (arguments.contains('--production')) {
      arguments.remove('--production');
      return true;
    }
    return false;
  }();

  final filePath = '${Directory.current.path}/bin/server.dart';
  final fileUri = new Uri.file(filePath);
  bool willRestart = true;
  SendPort restartPort;
  StreamController changeBroadcast = new StreamController.broadcast();

  final watcher = productionMode ? null : new Watcher(Directory.current.path).events.listen((event) {
    // if (!event.path.endsWith('.dart')) return;

    changeBroadcast.add(event);
    willRestart = true;
    restartPort?.send(null);
  });

  while (willRestart) {
    willRestart = false;
    final exitPort = new ReceivePort();
    final errorPort = new ReceivePort();
    final receiveExitCommandPort = new ReceivePort();
    final receiveInitialPortsSendPort = new ReceivePort();
    final childReloadPort = new ReceivePort();
    childReloadPort.listen((_) {
      willRestart = true;
      restartPort?.send(null);
    });
    await Isolate.spawnUri(
        fileUri,
        arguments,
        receiveInitialPortsSendPort.sendPort,
        onExit: exitPort.sendPort,
        onError: errorPort.sendPort,
        errorsAreFatal: false,
        automaticPackageResolution: true
    );
    final SendPort sendInitialPorts = await receiveInitialPortsSendPort.first;
    receiveInitialPortsSendPort.close();
    sendInitialPorts.send(receiveExitCommandPort.sendPort);

    restartPort = await receiveExitCommandPort.first
        .timeout(const Duration(seconds: 10));

    sendInitialPorts.send(childReloadPort.sendPort);

    final process = new Completer<int>();
    errorPort.listen((List l) async {
      print(l[0]);
      print('Listening for changes...');
      await changeBroadcast.stream.first;
      if (!process.isCompleted) process.complete(0);
    });
    exitPort.listen((_) {
      if (!process.isCompleted) process.complete(0);
    });

    exitCode = await process.future;
    childReloadPort.close();
    exitPort.close();
    errorPort.close();
    receiveExitCommandPort.close();
  }

  await watcher?.cancel();
}
