import 'package:embla/src/http/request_response.dart';
import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'package:embla/src/http/http_exceptions.dart';
import 'package:stack_trace/stack_trace.dart';
import 'package:embla/src/util/trace_formatting.dart';

class ErrorTemplate {
  static final _packagesPath = Platform.packageRoot ?? Directory.current.path + '/packages';

  static const String _hljs =
    r"""!function(e){var n="object"==typeof window&&window||"object"==typeof self&&self;"undefined"!=typeof exports?e(exports):n&&(n.hljs=e({}),"function"==typeof define&&define.amd&&define([],function(){return n.hljs}))}(function(e){function n(e){return e.replace(/[&<>]/gm,function(e){return I[e]})}function t(e){return e.nodeName.toLowerCase()}function r(e,n){var t=e&&e.exec(n);return t&&0===t.index}function a(e){return k.test(e)}function i(e){var n,t,r,i,o=e.className+" ";if(o+=e.parentNode?e.parentNode.className:"",t=B.exec(o))return R(t[1])?t[1]:"no-highlight";for(o=o.split(/\s+/),n=0,r=o.length;r>n;n++)if(i=o[n],a(i)||R(i))return i}function o(e,n){var t,r={};for(t in e)r[t]=e[t];if(n)for(t in n)r[t]=n[t];return r}function u(e){var n=[];return function r(e,a){for(var i=e.firstChild;i;i=i.nextSibling)3===i.nodeType?a+=i.nodeValue.length:1===i.nodeType&&(n.push({event:"start",offset:a,node:i}),a=r(i,a),t(i).match(/br|hr|img|input/)||n.push({event:"stop",offset:a,node:i}));return a}(e,0),n}function c(e,r,a){function i(){return e.length&&r.length?e[0].offset!==r[0].offset?e[0].offset<r[0].offset?e:r:"start"===r[0].event?e:r:e.length?e:r}function o(e){function r(e){return" "+e.nodeName+'="'+n(e.value)+'"'}l+="<"+t(e)+w.map.call(e.attributes,r).join("")+">"}function u(e){l+="</"+t(e)+">"}function c(e){("start"===e.event?o:u)(e.node)}for(var s=0,l="",f=[];e.length||r.length;){var g=i();if(l+=n(a.substr(s,g[0].offset-s)),s=g[0].offset,g===e){f.reverse().forEach(u);do c(g.splice(0,1)[0]),g=i();while(g===e&&g.length&&g[0].offset===s);f.reverse().forEach(o)}else"start"===g[0].event?f.push(g[0].node):f.pop(),c(g.splice(0,1)[0])}return l+n(a.substr(s))}function s(e){function n(e){return e&&e.source||e}function t(t,r){return new RegExp(n(t),"m"+(e.cI?"i":"")+(r?"g":""))}function r(a,i){if(!a.compiled){if(a.compiled=!0,a.k=a.k||a.bK,a.k){var u={},c=function(n,t){e.cI&&(t=t.toLowerCase()),t.split(" ").forEach(function(e){var t=e.split("|");u[t[0]]=[n,t[1]?Number(t[1]):1]})};"string"==typeof a.k?c("keyword",a.k):E(a.k).forEach(function(e){c(e,a.k[e])}),a.k=u}a.lR=t(a.l||/\w+/,!0),i&&(a.bK&&(a.b="\\b("+a.bK.split(" ").join("|")+")\\b"),a.b||(a.b=/\B|\b/),a.bR=t(a.b),a.e||a.eW||(a.e=/\B|\b/),a.e&&(a.eR=t(a.e)),a.tE=n(a.e)||"",a.eW&&i.tE&&(a.tE+=(a.e?"|":"")+i.tE)),a.i&&(a.iR=t(a.i)),null==a.r&&(a.r=1),a.c||(a.c=[]);var s=[];a.c.forEach(function(e){e.v?e.v.forEach(function(n){s.push(o(e,n))}):s.push("self"===e?a:e)}),a.c=s,a.c.forEach(function(e){r(e,a)}),a.starts&&r(a.starts,i);var l=a.c.map(function(e){return e.bK?"\\.?("+e.b+")\\.?":e.b}).concat([a.tE,a.i]).map(n).filter(Boolean);a.t=l.length?t(l.join("|"),!0):{exec:function(){return null}}}}r(e)}function l(e,t,a,i){function o(e,n){var t,a;for(t=0,a=n.c.length;a>t;t++)if(r(n.c[t].bR,e))return n.c[t]}function u(e,n){if(r(e.eR,n)){for(;e.endsParent&&e.parent;)e=e.parent;return e}return e.eW?u(e.parent,n):void 0}function c(e,n){return!a&&r(n.iR,e)}function g(e,n){var t=N.cI?n[0].toLowerCase():n[0];"""
    r"""return e.k.hasOwnProperty(t)&&e.k[t]}function h(e,n,t,r){var a=r?"":y.classPrefix,i='<span class="'+a,o=t?"":C;return i+=e+'">',i+n+o}function p(){var e,t,r,a;if(!E.k)return n(B);for(a="",t=0,E.lR.lastIndex=0,r=E.lR.exec(B);r;)a+=n(B.substr(t,r.index-t)),e=g(E,r),e?(M+=e[1],a+=h(e[0],n(r[0]))):a+=n(r[0]),t=E.lR.lastIndex,r=E.lR.exec(B);return a+n(B.substr(t))}function d(){var e="string"==typeof E.sL;if(e&&!x[E.sL])return n(B);var t=e?l(E.sL,B,!0,L[E.sL]):f(B,E.sL.length?E.sL:void 0);return E.r>0&&(M+=t.r),e&&(L[E.sL]=t.top),h(t.language,t.value,!1,!0)}function b(){k+=null!=E.sL?d():p(),B=""}function v(e){k+=e.cN?h(e.cN,"",!0):"",E=Object.create(e,{parent:{value:E}})}function m(e,n){if(B+=e,null==n)return b(),0;var t=o(n,E);if(t)return t.skip?B+=n:(t.eB&&(B+=n),b(),t.rB||t.eB||(B=n)),v(t,n),t.rB?0:n.length;var r=u(E,n);if(r){var a=E;a.skip?B+=n:(a.rE||a.eE||(B+=n),b(),a.eE&&(B=n));do E.cN&&(k+=C),E.skip||(M+=E.r),E=E.parent;while(E!==r.parent);return r.starts&&v(r.starts,""),a.rE?0:n.length}if(c(n,E))throw new Error('Illegal lexeme "'+n+'" for mode "'+(E.cN||"<unnamed>")+'"');return B+=n,n.length||1}var N=R(e);if(!N)throw new Error('Unknown language: "'+e+'"');s(N);var w,E=i||N,L={},k="";for(w=E;w!==N;w=w.parent)w.cN&&(k=h(w.cN,"",!0)+k);var B="",M=0;try{for(var I,j,O=0;;){if(E.t.lastIndex=O,I=E.t.exec(t),!I)break;j=m(t.substr(O,I.index-O),I[0]),O=I.index+j}for(m(t.substr(O)),w=E;w.parent;w=w.parent)w.cN&&(k+=C);return{r:M,value:k,language:e,top:E}}catch(T){if(T.message&&-1!==T.message.indexOf("Illegal"))return{r:0,value:n(t)};throw T}}function f(e,t){t=t||y.languages||E(x);var r={r:0,value:n(e)},a=r;return t.filter(R).forEach(function(n){var t=l(n,e,!1);t.language=n,t.r>a.r&&(a=t),t.r>r.r&&(a=r,r=t)}),a.language&&(r.second_best=a),r}function g(e){return y.tabReplace||y.useBR?e.replace(M,function(e,n){return y.useBR&&"\n"===e?"<br>":y.tabReplace?n.replace(/\t/g,y.tabReplace):void 0}):e}function h(e,n,t){var r=n?L[n]:t,a=[e.trim()];return e.match(/\bhljs\b/)||a.push("hljs"),-1===e.indexOf(r)&&a.push(r),a.join(" ").trim()}function p(e){var n,t,r,o,s,p=i(e);a(p)||(y.useBR?(n=document.createElementNS("http://www.w3.org/1999/xhtml","div"),n.innerHTML=e.innerHTML.replace(/\n/g,"").replace(/<br[ \/]*>/g,"\n")):n=e,s=n.textContent,r=p?l(p,s,!0):f(s),t=u(n),t.length&&(o=document.createElementNS("http://www.w3.org/1999/xhtml","div"),o.innerHTML=r.value,r.value=c(t,u(o),s)),r.value=g(r.value),e.innerHTML=r.value,e.className=h(e.className,p,r.language),e.result={language:r.language,re:r.r},r.second_best&&(e.second_best={language:r.second_best.language,re:r.second_best.r}))}function d(e){y=o(y,e)}function b(){if(!b.called){b.called=!0;var e=document.querySelectorAll("pre code");w.forEach.call(e,p)}}function v(){addEventListener("DOMContentLoaded",b,!1),addEventListener("load",b,!1)}function m(n,t){var r=x[n]=t(e);r.aliases&&r.aliases.forEach(function(e){L[e]=n})}function N(){return E(x)}function R(e){return e=(e||"").toLowerCase(),"""
    r"""x[e]||x[L[e]]}var w=[],E=Object.keys,x={},L={},k=/^(no-?highlight|plain|text)$/i,B=/\blang(?:uage)?-([\w-]+)\b/i,M=/((^(<[^>]+>|\t|)+|(?:\n)))/gm,C="</span>",y={classPrefix:"hljs-",tabReplace:null,useBR:!1,languages:void 0},I={"&":"&amp;","<":"&lt;",">":"&gt;"};return e.highlight=l,e.highlightAuto=f,e.fixMarkup=g,e.highlightBlock=p,e.configure=d,e.initHighlighting=b,e.initHighlightingOnLoad=v,e.registerLanguage=m,e.listLanguages=N,e.getLanguage=R,e.inherit=o,e.IR="[a-zA-Z]\\w*",e.UIR="[a-zA-Z_]\\w*",e.NR="\\b\\d+(\\.\\d+)?",e.CNR="(-?)(\\b0[xX][a-fA-F0-9]+|(\\b\\d+(\\.\\d*)?|\\.\\d+)([eE][-+]?\\d+)?)",e.BNR="\\b(0b[01]+)",e.RSR="!|!=|!==|%|%=|&|&&|&=|\\*|\\*=|\\+|\\+=|,|-|-=|/=|/|:|;|<<|<<=|<=|<|===|==|=|>>>=|>>=|>=|>>>|>>|>|\\?|\\[|\\{|\\(|\\^|\\^=|\\||\\|=|\\|\\||~",e.BE={b:"\\\\[\\s\\S]",r:0},e.ASM={cN:"string",b:"'",e:"'",i:"\\n",c:[e.BE]},e.QSM={cN:"string",b:'"',e:'"',i:"\\n",c:[e.BE]},e.PWM={b:/\b(a|an|the|are|I'm|isn't|don't|doesn't|won't|but|just|should|pretty|simply|enough|gonna|going|wtf|so|such|will|you|your|like)\b/},e.C=function(n,t,r){var a=e.inherit({cN:"comment",b:n,e:t,c:[]},r||{});return a.c.push(e.PWM),a.c.push({cN:"doctag",b:"(?:TODO|FIXME|NOTE|BUG|XXX):",r:0}),a},e.CLCM=e.C("//","$"),e.CBCM=e.C("/\\*","\\*/"),e.HCM=e.C("#","$"),e.NM={cN:"number",b:e.NR,r:0},e.CNM={cN:"number",b:e.CNR,r:0},e.BNM={cN:"number",b:e.BNR,r:0},e.CSSNM={cN:"number",b:e.NR+"(%|em|ex|ch|rem|vw|vh|vmin|vmax|cm|mm|in|pt|pc|px|deg|grad|rad|turn|s|ms|Hz|kHz|dpi|dpcm|dppx)?",r:0},e.RM={cN:"regexp",b:/\//,e:/\/[gimuy]*/,i:/\n/,c:[e.BE,{b:/\[/,e:/\]/,r:0,c:[e.BE]}]},e.TM={cN:"title",b:e.IR,r:0},e.UTM={cN:"title",b:e.UIR,r:0},e.METHOD_GUARD={b:"\\.\\s*"+e.UIR,r:0},e});hljs.registerLanguage("xml",function(s){var e="[A-Za-z0-9\\._:-]+",t={eW:!0,i:/</,r:0,c:[{cN:"attr",b:e,r:0},{b:/=\s*/,r:0,c:[{cN:"string",endsParent:!0,v:[{b:/"/,e:/"/},{b:/'/,e:/'/},{b:/[^\s"'=<>`]+/}]}]}]};return{aliases:["html","xhtml","rss","atom","xjb","xsd","xsl","plist"],cI:!0,c:[{cN:"meta",b:"<!DOCTYPE",e:">",r:10,c:[{b:"\\[",e:"\\]"}]},s.C("<!--","-->",{r:10}),{b:"<\\!\\[CDATA\\[",e:"\\]\\]>",r:10},{b:/<\?(php)?/,e:/\?>/,sL:"php",c:[{b:"/\\*",e:"\\*/",skip:!0}]},{cN:"tag",b:"<style(?=\\s|>|$)",e:">",k:{name:"style"},c:[t],starts:{e:"</style>",rE:!0,sL:["css","xml"]}},{cN:"tag",b:"<script(?=\\s|>|$)",e:">",k:{name:"script"},c:[t],starts:{e:"<\/script>",rE:!0,sL:["actionscript","javascript","handlebars","xml"]}},{cN:"meta",v:[{b:/<\?xml/,e:/\?>/,r:10},{b:/<\?\w+/,e:/\?>/}]},{cN:"tag",b:"</?",e:"/?>",c:[{cN:"name",b:/[^\/><\s]+/,r:0},t]}]}});hljs.registerLanguage("markdown",function(e){return{aliases:["md","mkdown","mkd"],c:[{cN:"section",v:[{b:"^#{1,6}",e:"$"},{b:"^.+?\\n[=-]{2,}$"}]},{b:"<",e:">",sL:"xml",r:0},{cN:"bullet",b:"^([*+-]|(\\d+\\.))\\s+"},{cN:"strong",b:"[*_]{2}.+?[*_]{2}"},{cN:"emphasis",v:[{b:"\\*.+?\\*"},{b:"_.+?_",r:0}]},{cN:"quote",b:"^>\\s+",e:"$"},{cN:"code",v:[{b:"^```w*s*$",e:"^```s*$"},{b:"`.+?`"},{b:"^( {4}|	)",e:"$",r:0}]},"""
    r"""{b:"^[-\\*]{3,}",e:"$"},{b:"\\[.+?\\][\\(\\[].*?[\\)\\]]",rB:!0,c:[{cN:"string",b:"\\[",e:"\\]",eB:!0,rE:!0,r:0},{cN:"link",b:"\\]\\(",e:"\\)",eB:!0,eE:!0},{cN:"symbol",b:"\\]\\[",e:"\\]",eB:!0,eE:!0}],r:10},{b:/^\[[^\n]+\]:/,rB:!0,c:[{cN:"symbol",b:/\[/,e:/\]/,eB:!0,eE:!0},{cN:"link",b:/:\s*/,e:/$/,eB:!0}]}]}});hljs.registerLanguage("dart",function(e){var t={cN:"subst",b:"\\$\\{",e:"}",k:"true false null this is new super"},r={cN:"string",v:[{b:"r'''",e:"'''"},{b:'r"""
    '"""'
    "',e:'"
    '"""'
    r"""'},{b:"r'",e:"'",i:"\\n"},{b:'r"',e:'"',i:"\\n"},{b:"'''",e:"'''",c:[e.BE,t]},{b:'"""
    '"""'
    r"""',e:'"""
    '"""'
    r"""',c:[e.BE,t]},{b:"'",e:"'",i:"\\n",c:[e.BE,t]},{b:'"',e:'"',i:"\\n",c:[e.BE,t]}]};t.c=[e.CNM,r];var n={keyword:"assert async await break case catch class const continue default do else enum extends false final finally for if in is new null rethrow return super switch sync this throw true try var void while with yield abstract as dynamic export external factory get implements import library operator part set static typedef",built_in:"print Comparable DateTime Duration Function Iterable Iterator List Map Match Null Object Pattern RegExp Set Stopwatch String StringBuffer StringSink Symbol Type Uri bool double int num document window querySelector querySelectorAll Element ElementList"};return{k:n,c:[r,e.C("/\\*\\*","\\*/",{sL:"markdown"}),e.C("///","$",{sL:"markdown"}),e.CLCM,e.CBCM,{cN:"class",bK:"class interface",e:"{",eE:!0,c:[{bK:"extends implements"},e.UTM]},e.CNM,{cN:"meta",b:"@[A-Za-z]+"},{b:"=>"}]}});""";

  static const String _style = r"""
/* system-font.css v1.1.0 | CC0-1.0 License | github.com/jonathantneal/system-font-face */
@font-face {
  font-family: system;
  font-style: normal;
  font-weight: 300;
  src: local(".SFNSText-Light"), local(".HelveticaNeueDeskInterface-Light"), local(".LucidaGrandeUI"), local("Ubuntu Light"), local("Segoe UI Light"), local("Roboto-Light"), local("DroidSans"), local("Tahoma");
}
@font-face {
  font-family: system;
  font-style: italic;
  font-weight: 300;
  src: local(".SFNSText-LightItalic"), local(".HelveticaNeueDeskInterface-Italic"), local(".LucidaGrandeUI"), local("Ubuntu Light Italic"), local("Segoe UI Light Italic"), local("Roboto-LightItalic"), local("DroidSans"), local("Tahoma");
}
@font-face {
  font-family: system;
  font-style: normal;
  font-weight: 400;
  src: local(".SFNSText-Regular"), local(".HelveticaNeueDeskInterface-Regular"), local(".LucidaGrandeUI"), local("Ubuntu"), local("Segoe UI"), local("Roboto-Regular"), local("DroidSans"), local("Tahoma");
}
@font-face {
  font-family: system;
  font-style: italic;
  font-weight: 400;
  src: local(".SFNSText-Italic"), local(".HelveticaNeueDeskInterface-Italic"), local(".LucidaGrandeUI"), local("Ubuntu Italic"), local("Segoe UI Italic"), local("Roboto-Italic"), local("DroidSans"), local("Tahoma");
}
@font-face {
  font-family: system;
  font-style: normal;
  font-weight: 500;
  src: local(".SFNSText-Medium"), local(".HelveticaNeueDeskInterface-MediumP4"), local(".LucidaGrandeUI"), local("Ubuntu Medium"), local("Segoe UI Semibold"), local("Roboto-Medium"), local("DroidSans-Bold"), local("Tahoma Bold");
}
@font-face {
  font-family: system;
  font-style: italic;
  font-weight: 500;
  src: local(".SFNSText-MediumItalic"), local(".HelveticaNeueDeskInterface-MediumItalicP4"), local(".LucidaGrandeUI"), local("Ubuntu Medium Italic"), local("Segoe UI Semibold Italic"), local("Roboto-MediumItalic"), local("DroidSans-Bold"), local("Tahoma Bold");
}
@font-face {
  font-family: system;
  font-style: normal;
  font-weight: 700;
  src: local(".SFNSText-Bold"), local(".HelveticaNeueDeskInterface-Bold"), local(".LucidaGrandeUI"), local("Ubuntu Bold"), local("Roboto-Bold"), local("DroidSans-Bold"), local("Segoe UI Bold"), local("Tahoma Bold");
}
@font-face {
  font-family: system;
  font-style: italic;
  font-weight: 700;
  src: local(".SFNSText-BoldItalic"), local(".HelveticaNeueDeskInterface-BoldItalic"), local(".LucidaGrandeUI"), local("Ubuntu Bold Italic"), local("Roboto-BoldItalic"), local("DroidSans-Bold"), local("Segoe UI Bold Italic"), local("Tahoma Bold");
}
body {
  background: #f8f8f8;
  margin: 0;
  font-family: system, sans-serif;
  color: #333;
}
.container {
  background: #fefefe;
  display: flex;
  flex-direction: column;
  max-width: 130em;
  margin: auto;
  height: 100vh;
  box-shadow: 0 -5em 6em -5em #000;
}
.container__header {
  padding: 1rem;
}
.container__header__code {
  color: #555;
  font-size: 0.9em;
  font-family: 'SF Mono', 'Menlo', monospace;
}
.container__header__heading {
  margin: 0.4em 0 0;
  font-weight: 300;
}
.container__content {
  flex: 1;
  display: flex;
}
.source {
  flex: 2 2 66%;
  display: flex;
}
.source__pre {
  overflow: auto;
  padding: 1em;
  border-top: 1px solid transparentize(#333, 0.9);
  line-height: 1.4;
  font-size: 0.9em;
  margin: 0;
  background: #f8f8f8;
  flex: 1;
}
.source__pre,
.source__pre code {
  font-family: 'SF Mono', 'Menlo', monospace;
}
.source__pre:not(.active) {
  display: none;
}
.source__highlighted-line {
  border-bottom: 2px solid #de4530;
}
.source__line-number .hljs-number {
  color: #888;
}
.stack {
  flex: 1 1 33%;
  overflow: auto;
}
.stack__list {
  margin: 0;
  list-style: none;
  padding: 0;
  font-family: 'SF Mono', 'Menlo', monospace;
  cursor: pointer;
}
.stack__list__item {
  width: 100%;
  border: 0;
  background: #fefefe;
  margin: 0;
  font: inherit;
  text-align: left;
  -webkit-appearance: none;
  -moz-appearance: none;
  -ms-appearance: none;
  -o-appearance: none;
  appearance: none;
  padding: 1em;
  cursor: pointer;
  border-top: 1px solid transparentize(#333, 0.9);
}
.stack__list__item:focus {
  outline: 0;
  box-shadow: inset 0 0 0 2px #798598;
}
.stack__list__item__file {
  overflow: hidden;
  white-space: nowrap;
  text-overflow: ellipsis;
  position: relative;
  display: block;
  margin-bottom: 0.2em;
}
.active .stack__list__item {
  border: 1px solid transparentize(#333, 0.9);
  background-color: #798598;
  color: #fefefe;
}
.active .stack__list__item:active {
  color: #fefefe;
}
.active + li .stack__list__item {
  border-top: 0;
}
.stack__list__item:active {
  color: #555;
}
.stack__list__item__function {
  margin: 0;
}
/**
 * GitHub Gist Theme
 * Author : Louis Barranqueiro - https://github.com/LouisBarranqueiro
 */

.hljs {
  display: block;
  /* background: white; */
  padding: 0.5em;
  color: #333333;
  overflow-x: auto;
}

.hljs-comment,
.hljs-meta {
  color: #969896;
}

.hljs-string,
.hljs-variable,
.hljs-template-variable,
.hljs-strong,
.hljs-emphasis,
.hljs-quote {
  color: #df5000;
}

.hljs-keyword,
.hljs-selector-tag,
.hljs-type {
  color: #a71d5d;
}

.hljs-literal,
.hljs-symbol,
.hljs-bullet,
.hljs-attribute {
  color: #0086b3;
}

.hljs-section,
.hljs-name {
  color: #63a35c;
}

.hljs-tag {
  color: #333333;
}

.hljs-title,
.hljs-attr,
.hljs-selector-id,
.hljs-selector-class,
.hljs-selector-attr,
.hljs-selector-pseudo {
  color: #795da3;
}

.hljs-addition {
  color: #55a532;
  background-color: #eaffea;
}

.hljs-deletion {
  color: #bd2c00;
  background-color: #ffecec;
}

.hljs-link {
  text-decoration: underline;
}
  """;

  ErrorTemplate();

  Future<Response> catchError(error, StackTrace stack) async {
    final code = _code(error);
    return new Response(
      code,
      body: renderError(error, stack, code),
      headers: {
        'Content-Type': ContentType.HTML.toString()
      }
    );
  }

  Stream<List<int>> renderError(error, StackTrace stack, int code) {
    return _template(error, new Chain.forTrace(stack), code).map(UTF8.encode);
  }

  int _code(error) {
    if (error is HttpBadRequestException) return 400;
    if (error is HttpUnauthorizedException) return 401;
    if (error is HttpPaymentRequiredException) return 402;
    if (error is HttpForbiddenException) return 403;
    if (error is HttpNotFoundException) return 404;
    if (error is HttpMethodNotAllowedException) return 405;
    if (error is HttpNotAcceptableException) return 406;
    if (error is HttpProxyAuthenticationRequiredException) return 407;
    if (error is HttpRequestTimeoutException) return 408;
    if (error is HttpConflictException) return 409;
    if (error is HttpGoneException) return 410;
    if (error is HttpLengthRequiredException) return 411;
    if (error is HttpPreconditionFailedException) return 412;
    if (error is HttpPayloadTooLargeException) return 413;
    if (error is HttpURITooLongException) return 414;
    if (error is HttpUnsupportedMediaTypeException) return 415;
    if (error is HttpRangeNotSatisfiableException) return 416;
    if (error is HttpExpectationFailedException) return 417;
    if (error is HttpImATeapotException) return 418;
    if (error is HttpAuthenticationTimeoutException) return 419;
    if (error is HttpMisdirectedRequestException) return 421;
    if (error is HttpUnprocessableEntityException) return 422;
    if (error is HttpLockedException) return 423;
    if (error is HttpFailedDependencyException) return 424;
    if (error is HttpUpgradeRequiredException) return 426;
    if (error is HttpPreconditionRequiredException) return 428;
    if (error is HttpTooManyRequestsException) return 429;
    if (error is HttpRequestHeaderFieldsTooLargeException) return 431;
    if (error is HttpInternalServerErrorException) return 500;
    if (error is HttpNotImplementedException) return 501;
    if (error is HttpBadGatewayException) return 502;
    if (error is HttpServiceUnavailableException) return 503;
    if (error is HttpGatewayTimeoutException) return 504;
    if (error is HttpVersionNotSupportedException) return 505;
    if (error is HttpVariantAlsoNegotiatesException) return 506;
    if (error is HttpInsufficientStorageException) return 507;
    if (error is HttpLoopDetectedException) return 508;
    if (error is HttpNotExtendedException) return 510;
    if (error is HttpNetworkAuthenticationRequiredException) return 511;
    return 500;
  }

  String _esc(input) {
    return input.toString()
      .replaceAll('<', '&lt;')
      .replaceAll('>', '&gt;');
  }

  Uri _resolveUri(Uri input) {
    if (input.scheme == 'dart' || !input.path.endsWith('.dart')) {
      return null;
    }
    if (input.scheme == 'package') {
      return Uri.parse('$_packagesPath/${input.path}');
    }
    return input;
  }

  Stream<String> _template(error, Chain chain, int statusCode) async* {
    final formatter = new TraceFormatter(chain);
    final List<Frame> frames = formatter.frames.reversed.toList();
    final String message = error.toString() == "Instance of '${error.runtimeType}'"
      ? error.runtimeType.toString()
      : error.toString();

    yield '''
<!DOCTYPE html>
<html lang="en">
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width">
  <title>$statusCode – ${_esc(message)}</title>
  <style>$_style</style>
  <script>$_hljs</script>
  <script>hljs.initHighlightingOnLoad();</script>
</head>
<body>
  <div class="container">
    <header class="container__header">
      <small class="container__header__code">Error $statusCode</small>
      <h1 class="container__header__heading">${_esc(message)}</h1>
    </header>
    <div class="container__content">
      <aside class="stack">
        <ul class="stack__list">
          ${frames.map((Frame frame) {
            final index = frames.indexOf(frame);
            return """<li class="${index == 0 ? 'active' : ''}">
              <button class="stack__list__item" id="stack-frame-$index">
                <small class="stack__list__item__file">
                  ${_esc(frame.uri.path.replaceFirst(Directory.current.path+'/', ''))}
                </small>
                <h5 class="stack__list__item__function">
                  ${_esc(frame.member)}
                </h5>
              </button>
            </li>""";
          }).join()}
        </ul>
      </aside>
      <main class="source">''';

        for (final Frame frame in frames) {
          final index = frames.indexOf(frame);
          final uri = _resolveUri(frame.uri);
          final Stream<String> file = uri == null ? new Stream.empty() : new File.fromUri(uri).openRead().map(UTF8.decode).map((s) => _esc(s));
          yield '<pre class="source__pre ${index == 0 ? 'active' : ''}" id="source-for-stack-frame-$index"><code class="dart">';
          var line = 0;
          final lines = (await file.join()).split('\n');
          final lineColumnWidth = lines.length.toString().length + 1;
          yield lines.map((s) {
            line++;
            final linePrefix = '<span class="source__line-number">${line.toString().padRight(lineColumnWidth)}</span>';
            if (line == frame.line) {
              s = s.replaceFirstMapped(new RegExp('^(\\s*)'), (m) => m[1] + '<span class="source__highlighted-line">');
              return '$linePrefix $s</span>';
            }
            return '$linePrefix $s';
          }).join('\n');
          yield '</code></pre>';
        }

    yield '''
      </main>
    </div>
  </div>

  <script>
    var items = Array.prototype.slice.call(document.querySelectorAll(".stack__list__item"));
    var activeItem = document.querySelector(".stack__list .active");
    var sources = Array.prototype.slice.call(document.querySelectorAll(".source__pre"))
    var activeSource = document.querySelector(".source__pre.active");
    centerHighlight();
    items.forEach(function(i) {
      i.addEventListener("click", function() {
        removeActive(activeItem);
        removeActive(activeSource);
        activeItem = i.parentNode;
        activeSource = document.querySelector("#source-for-" + i.id);
        addActive(activeItem);
        addActive(activeSource);
        centerHighlight();
      });
    });

    function centerHighlight() {
      var highlight;
      if (highlight = activeSource.querySelector('.source__highlighted-line')) {
        activeSource.scrollTop = highlight.getBoundingClientRect().top + activeSource.scrollTop - 200;
      } else {
        activeSource.scrollTop = 0;
      }
    }

    function removeActive(element) {
      element.className = element.className.replace(/active/, '');
    }

    function addActive(element) {
      element.className += " active";
    }
  </script>
</body>
</html>

    ''';
  }
}
